#include <stdio.h>

#define MAX_TUNES 20
#define MAX_SONGS 64
#define MAX_PATTERNS 64

unsigned char mem[0x2000];
unsigned int sb_base = 0xA000;
long song_base = 0x70000L;
unsigned int patn_base = 0x0;

char *notes[12] = { "C", "Db", "D", "Eb", "E", "F", "Gb", "G", "Ab", "A", "Bb", "B" };

unsigned char get_mem(unsigned int addr)

{
	return mem[addr-sb_base];
}

unsigned int pattern_addr[MAX_PATTERNS];
unsigned int next_free_pattern = 0;

int add_pattern(unsigned int addr)
{
	int i;
	// scan for existing pattern
	for (i=0; i<next_free_pattern; i++)
		if (pattern_addr[i] == addr) return i;
	if (next_free_pattern < MAX_PATTERNS)
		pattern_addr[next_free_pattern++] = addr;
	else
		i=-1;
	return i;
}

unsigned int tune_addr[MAX_TUNES];
unsigned int next_free_tune = 0;
unsigned int song_addr[MAX_SONGS];
unsigned int next_free_song = 0;

int add_tune(unsigned int addr)
{
	int i;
	// scan for existing song
	for (i=0; i<next_free_tune; i++)
		if (tune_addr[i] == addr) return i;
	if (next_free_tune < MAX_TUNES)
		tune_addr[next_free_tune++] = addr;
	else
		i=-1;
	return i;
}


int add_song(unsigned int addr)
{
	int i;
	// scan for existing song
	for (i=0; i<next_free_song; i++)
		if (song_addr[i] == addr) return i;
	if (next_free_song < MAX_SONGS)
		song_addr[next_free_song++] = addr;
	else
		i=-1;
	return i;
}

void address_arg(unsigned int addr_lo, unsigned int addr_hi)
{
	printf("0x%04X", addr_lo + 256 * addr_hi);
}

void single_arg(unsigned int addr)
{
	printf("0x%02X", get_mem(addr));
}

void duration(unsigned int dur)
{
	printf("0x%02X\n", dur);
}

void parse_cmds(unsigned int addr)
{
	unsigned char cmd;
	do
	{
		cmd = get_mem(addr++);

		if (cmd < 0xC0)
		{
			printf("\tPLAY_NOTE %s, ", notes[cmd >> 4]);
			duration(cmd & 0xF);
		}
		else if (cmd < 0xD0)
		{
			printf("\tPLAY_REST ");
			duration(cmd & 0xF);
		}
		else if (cmd < 0xE0)
		{
			printf("\tCMD%02X\n", cmd);
		}
		else
		{
			switch (cmd & 0x1F)
			{
			case 0:
				printf("\tCMD0 0x%02X\n", get_mem(addr++));
				break;
			case 1:
			case 2:
			case 3:
			case 4:
			case 5:
			case 6:
			case 7:
			case 8:
			case 9:
			case 10:
			case 11:
			case 12:
			case 13:
			case 14:
				printf("\tSET_VOLUME 0x%02X\n", (cmd & 0xF) + 1);
				break;
			case 15:
			case 16:
			case 17:
			case 18:
			case 19:
			case 20:
				printf("\tSET_OCTAVE 0x%02X\n", (cmd & 0x1F) - 15);
				break;
			case 21:
				printf("\tCMD23 0x30, ");
				single_arg(addr++);
				printf(", ");
				single_arg(addr++);
				printf("\n");
				break;
			case 22:
				printf("\tCMD23 0x70, ");
				single_arg(addr++);
				printf(", ");
				single_arg(addr++);
				printf("\n");
				break;
			case 23:
				printf("\tCMD23 0xB0, ");
				single_arg(addr++);
				printf(", ");
				single_arg(addr++);
				printf("\n");
				break;
			case 24:
				printf("\tSET_SQ1_SWEEP ");
				single_arg(addr++);
				printf("\n");
				break;
			case 25:
				printf("\tSET_SOMETHING_OFF\n");
				break;
			case 26:
				printf("\tSET_SOMETHING_ON\n");
				break;
			case 27:
				printf("\tCMD27 ");
				single_arg(addr++);
				printf("\n");
				break;
			case 28:
				printf("\tCMD28_ADDR ");
				address_arg(get_mem(addr++), get_mem(addr++));
				printf("\n");
				break;
			case 29:
				printf("\tCMD29_ADDR ");
				address_arg(get_mem(addr++), get_mem(addr++));
				printf("\n");
				break;
			case 30:
				printf("\tGOTO_ADDR ");
				address_arg(get_mem(addr++), get_mem(addr++));
				printf("\n");
				break;
			case 31:
				printf("%dHALT_PLAYER\n", cmd & 0xF);
				break;
			}
		}
	} while (cmd < 0xFE);
}

int main(int argc, char *argv[])
{
	FILE *fp = fopen("FF3.mem","rb");
	int tune, song, pattern;
	unsigned int tune_ptr;
	unsigned int song_ptr;

	fseek(fp, song_base, SEEK_SET);
	fread(mem, 1, 0x2000, fp);

//	printf("\t.include \"notes.inc\"\n\t.include \"ag.inc\"\n");
	printf("NoSong = 0xFFFF\n\n");
	printf("\t.data\n\t.export TUNE_TAB\nTUNE_TAB:\n");

	tune_ptr = sb_base;
	for (tune=0, song_ptr = get_mem(tune_ptr++)+256* get_mem(tune_ptr++);
	     (tune==0 || ((tune>0) && (tune_ptr != tune_addr[0]))) && next_free_tune < MAX_TUNES;
	     tune++, song_ptr = get_mem(tune_ptr++)+256* get_mem(tune_ptr++))
	{
		printf("\t.word Tune%d ; 0x%04X\n", add_tune(song_ptr)+1, song_ptr);
	}

	for (tune=0; tune<next_free_tune; tune++)
	{
		tune_ptr = tune_addr[tune];
		printf("\nTune%d ; 0x%04X\n", tune, tune_ptr);
		for (song=0; song<5; song++)
		{
			song_ptr = get_mem(tune_ptr++)+256* get_mem(tune_ptr++);
			if (song_ptr == 0xFFFF)
			{
				printf("\t.word NoSong\n");
			}
			else
			{
				printf("\t.word Song%d ; 0x%04X\n", add_song(song_ptr)+1, song_ptr);
			}
		}
	}

	for (song=0; song<next_free_song; song++)
	{
		patn_base = song_addr[song];
		printf("\nSong%d: ; 0x%04X\n", song + 1, patn_base);
		parse_cmds(patn_base);
	}
}
